#!/usr/bin/perl -w

# LSB Test Execution Framework
# Tests Execution Module (tests_exec.pl)
#
# Copyright (C) 2006-2007 Free Standards Group. All rights reserved.
#
# This program has been developed by ISP RAS for FSG.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

use Templates;
use POSIX ":sys_wait_h";
use Fcntl ':flock';

autoflush_off();

sub is_process_finished($) {
	my $PID = $_[0];
	my $test_line = `ps -p $PID -o s --no-heading`;
	$test_line =~ s/[\r\n]//g;
	return ($test_line !~ m/^[^Zz]/);
}

my $script_begin = "<script language=\"javascript\" type=\"text/javascript\">\n// <![CDATA[\n";
my $script_end = "\n// ]]>\n</script>\n";

print "HTTP/1.0 200 OK\n";
print "Content-type: text/html\n\n";

print_header("Test Execution Console Page", "exec");

print <<DATA;
<h1>Console output</h1>
<form method="get" action="" name="select_test">
Please, select which profile to run:&nbsp;&nbsp;&nbsp;
DATA

my $found = 0;
my $profiles_list = "";
if (opendir(DIR, "$DOCUMENT_ROOT/profiles")) {
	my @files = sort grep !/^[\.~]/, readdir(DIR);
	foreach (@files) {
		if ($_GET{"profile"} and ($_GET{"profile"} eq $_)) {
			$profiles_list .= "    <option value=\"$_\" selected=\"selected\">$_</option>\n";
		}
		else {
			$profiles_list .= "    <option value=\"$_\">$_</option>\n";
		}
		$found = 1;
	}
	closedir(DIR);
}

if ($found) {
	print <<DATA;
  <select name="profile" id="test_profile" style="width: 150px;">
$profiles_list  </select>&nbsp;
  <input type="button" name="Start" id="start_button" value="Start" style="width: 80px;" onclick="javascript:document.select_test.submit();" />&nbsp;
  <input type="button" name="Stop" id="stop_button" value="Stop" style="width: 80px;" onclick="javascript:stop_test();" disabled="disabled" />
DATA
}
else {
	print <<DATA;
  <select name="profile" id="test_profile_no" style="width: 150px;" disabled="disabled">
    <option value="">&lt;No profiles present&gt;</option>
  </select>&nbsp;
  <input type="button" name="Start" id="start_button_no" value="Start" style="width: 80px;" disabled="disabled" />&nbsp;
  <input type="button" name="Stop" id="stop_button" value="Stop" style="width: 80px;" onclick="javascript:stop_test();" disabled="disabled" />
DATA
}

if ($_GET{"cert"}) {
	print "  <input type=\"hidden\" name=\"cert\" value=\"$_GET{cert}\" />\n";
}

print <<DATA;
</form>
<br /><b><font color="blue"><span id="execInfo"></span></font> <span id="execStatus"></span></b><br />
<pre id="cmdlog" style="height: 400px; width: 700px; overflow: auto; text-wrap: none; border: 1px solid black; border-color: black; background-color: black; color: lightgreen;"></pre>
<div style="width: 700px; text-align: right;"><b><a id="view_report" style="display: none;" href="tests_results.pl">View test report</a></b></div>
<script language="javascript" type="text/javascript">
// <![CDATA[
var lastUpdateTime = new Date().getTime();
var updateBuff = "";
var lastLine = "";
var execFinished = false;
var inProgress = false;
var refreshDelay = 1000;

var browser = navigator.appName;

function refresh(force)
{
	if (!force && !inProgress)
	{
		var newTime = new Date().getTime();
		if (newTime - lastUpdateTime > refreshDelay)
			force = true;
	}

	if (execFinished)
		force = true;

	if (force)
	{
		inProgress = true;
		var refreshStartTime = new Date().getTime();
		var obj = document.getElementById("cmdlog");
		var height;
		if (browser.indexOf("Opera") != -1)
			height = obj.style.pixelHeight;
		else
			height = obj.clientHeight;
		var scroll_top = obj.scrollTop;
		var scroll_height = obj.scrollHeight;
		var auto_scroll = (scroll_height - height - scroll_top < 10);
		if (typeof(obj.innerText) == "undefined")
			obj.textContent = updateBuff + lastLine;
		else
			obj.innerText = updateBuff + lastLine;
		if (auto_scroll) {
			obj.scrollTop = obj.scrollHeight;
			obj.scrollTop = obj.scrollHeight;		// IE fix
		}
		lastUpdateTime = new Date().getTime();
		var rTime = (lastUpdateTime - refreshStartTime);
		inProgress = false;
	}
}

var re_x08_m = new RegExp();
var re_x08_b = new RegExp();
re_x08_m.compile("[^\\x08]\\x08", "g");
re_x08_b.compile("^\\x08+", "g");

function processSpecChars(line) {
	var res = line;
	var tmp;
	while (1) {
		tmp = res.replace(re_x08_m, "");
		if (tmp != res)
			res = tmp;
		else
			break;
	}
	res = res.replace(re_x08_b, "");
	return res;
}

function ao(txt)
{
	lastLine += txt;
	var pos1 = 0;
	var pos2;
	var firstLine = true;
	do {
		pos2 = lastLine.indexOf("\\n", pos1);
		var currentLine;
		if (pos2 != -1) {
			currentLine = lastLine.substring(pos1, pos2 + 1);
			if (firstLine) {
				currentLine = processSpecChars(currentLine);
				firstLine = false;
			}
			updateBuff += currentLine;
			pos1 = pos2 + 1;
		}
		else {
			lastLine = lastLine.substring(pos1);
			lastLine = processSpecChars(lastLine);
		}
	} while (pos2 != -1);

	refresh(false);
}

function setInfo(txt)
{
	var obj = document.getElementById("execInfo");
	obj.innerHTML = txt;
}

function setStatus(txt)
{
	var obj = document.getElementById("execStatus");
	obj.innerHTML = " " + txt;
}

function timerfunc()
{
	refresh();
	if (!execFinished)
		setTimeout("timerfunc()", refreshDelay);
}
DATA

my $save_profile_name = "";
if ($_GET{"profile"} and ($_GET{"profile"} !~ m/^~/)) {
	$save_profile_name = "profile=".$_GET{"profile"}."&";
}

print <<DATA;
var kill_pid = 0;
function stop_test() {
	if (confirm("Do you really want to kill the process (PID: " + kill_pid + ")?")) {
		var new_url = "tests_kill.pl?" + "$save_profile_name" + "exec=1&pid=" + kill_pid;
		document.location = new_url;
	}
}

setTimeout("timerfunc()", refreshDelay);
// ]]>
</script>
DATA

sub processSpecChars($) {
	my $tmp = $_[0];
	while (1) {
		$tmp =~ s/[^\x08]\x08//sg;
		if ($tmp ne $_[0]) {
			$_[0] = $tmp;
		}
		else {
			last;
		}
	}
	$_[0] =~ s/^\x08+//s;
}

sub process_output($) {
	$_[0] =~ s/\r//g;
	my $pos1 = 0;
	my $pos2;
	my $firstLine = 1;
	my $buffer = "";
	while(1) {
		$pos2 = index($_[0], "\n", $pos1);
		my $currentLine;
		if ($pos2 != -1) {
			$currentLine = substr($_[0], $pos1, $pos2 - $pos1 + 1);
			if (!$firstLine) {
				processSpecChars($currentLine);
			}
			else {
				$firstLine = 0;
			}
			$buffer .= $currentLine;
			$pos1 = $pos2 + 1;
		}
		else {
			$buffer .= substr($_[0], $pos1);
			last;
		}
	}

	$_[0] = $buffer;
	$_[0] =~ s/([\'\"\\])/\\$1/g;
	$_[0] =~ s/\n/\\n/g;
}

sub do_update_output(@) {
	my ($PID, $kill_PID, $file_name, $main_reader, $FILE) = @_;

	print $script_begin."setInfo('Tests are running...');\n";
	print "var elem = document.getElementById(\"test_profile\");\nif (elem) { elem.disabled = 1; }\n";
	print "elem = document.getElementById(\"start_button\");\nif (elem) { elem.disabled = 1; }\n";
	print "kill_pid = $kill_PID;\nelem = document.getElementById(\"stop_button\");\nif (elem) { elem.disabled = 0; }".$script_end;
	my $buffer;
	while (1) {
		my $finished = is_process_finished($PID);
		my $sz = -s $file_name;						# $sz == File size
		$sz -= tell($FILE);							# $sz == Bytes to read
		while($sz) {								# Read by 4096-bytes portions
			my $received = read($FILE, $buffer, ($sz > 4096) ? 4096 : $sz);
			if ($received == 0) {
				last;
			}
			$sz -= $received;
			process_output($buffer);
			print $script_begin."ao('$buffer');".$script_end;
	    }
		if (!$finished) {
			sleep(1);
		}
		else {
			my $report_dir = "";
			if (open(FILE_STAT, "< /tmp/tests_status.8888")) {
				my $status = <FILE_STAT>;
				$report_dir = <FILE_STAT>;
				close(FILE_STAT);
				print $script_begin;
				print "setInfo('Tests execution finished.');\n";
				$report_dir =~ s/[\r\n]//g if ($report_dir);
				my $status_txt = "";
				if ($status) {
					$status =~ s/[\r\n]//g;
					my $sig_code = $status & 127;
					$status = $status >> 8;
					if ($status == 0) {
						if ($sig_code != 0) {
							$status_txt = "Status: <font color=\"red\">Process killed by signal $sig_code!</font>";
						}
					}
					else {
						$status_txt = "Status: <font color=\"red\">Error (returned code: $status)!</font>";
					}
				}
				else {
					$status_txt = "<font color=\"#ff7f00\">Warning: Could not obtain return code!</font>";
				}
				if ($status_txt eq "") {
					if ($report_dir) {
						if (open(FILE_INFO, $CONFIG{"RESULTS_DIR"}."/$report_dir/INFO")) {
							while (<FILE_INFO>) {
								s/[\r\n]//g;
								if (m/^Verdict:\s*(.*)/) {
									if ($1 eq "Success") {
										$status_txt = "Status: <font color=\"green\">Success.</font>";
									}
									elsif ($1 eq "Failed") {
										$status_txt = "Status: <font color=\"red\">Failed.</font>";
									}
									last;
								}
							}
							close(FILE_INFO);
						}
					}
				}
				if ($status_txt eq "") {
					$status_txt = "Status: <font color=\"#ff7f00\">Unknown.</font>";
				}
				print "setStatus('$status_txt');";
				if ($report_dir) {
					my $cert = "";
					if ($_GET{"cert"}) {
						$cert = "&cert=".$_GET{"cert"};
					}
					elsif (-f $CONFIG{"RESULTS_DIR"}."/$report_dir/cert.html") {
						$cert = "&cert=1";
					}
					print "var elem = document.getElementById('view_report');\nif (elem) {\n    elem.style.display = '';\n    elem.setAttribute('href', 'tests_results.pl?details=$report_dir$cert');\n}";
					if ($_GET{"cert"}) {
						print "\ndocument.location = 'tests_results.pl?details=$report_dir$cert';";
					}
				}
				print $script_end;
			}
			else {
				print $script_begin."setInfo('Tests execution finished.');\nsetStatus('<font color=\"#ff7f00\">Warning: Could not obtain return code!</font>');".$script_end;
			}
			my $to_remove;
			if ($main_reader) {
				flock($FILE, LOCK_EX);
				$to_remove = $report_dir ? 0 : 1;
				if (!$to_remove) {
					my $dir_name = $CONFIG{"RESULTS_DIR"}."/".$report_dir."/autotest_info";
					mkdir($dir_name) unless (-d $dir_name);
					$to_remove = rename("/tmp/tests_output.8888", $dir_name."/tests_output.8888") ? 0 : 1;
				}
			}
			else {
				$to_remove = flock($FILE, LOCK_EX | LOCK_NB);
			}
			if ($to_remove) {
#				unlink("/tmp/tests_output.8888");
				rename("/tmp/tests_output.8888", "/tmp/tests_output.8888.old");
			}
			flock($FILE, LOCK_UN);
			close($FILE);
			last;
		}
	}
	print $script_begin."var elem = document.getElementById(\"test_profile\");\nif (elem) { elem.disabled = 0; }\n";
	print "elem = document.getElementById(\"start_button\");\nif (elem) { elem.disabled = 0; }\n";
	print "elem = document.getElementById(\"stop_button\");\nif (elem) { elem.disabled = 1; }".$script_end;
}

print_footer(1);     # do not finish with </body></html> now!

my $file_name;
if ($_GET{"killed"}) {
	$file_name = "/tmp/tests_output.8888.old";
	select(undef, undef, undef, 0.5);	# Sleep a bit to allow another process rename the file.
}
else {
	$file_name = "/tmp/tests_output.8888";
}

my $FILE;
if (open($FILE, "< $file_name")) {
	flock($FILE, LOCK_SH);
	my $PID = <$FILE>;
	my $kill_PID = $PID;
	$PID =~ s/PID: (\d+),.*/$1/g;
	$PID =~ s/[\r\n]//g;
	$kill_PID =~ s/.*CHLD_PID: (\d+).*/$1/g;
	$kill_PID =~ s/[\r\n]//g;

	do_update_output($PID, $kill_PID, $file_name, 0, $FILE);
}
elsif (!$_GET{"killed"}) {
	if ($_GET{"profile"}) {
		my $profile_path = $DOCUMENT_ROOT."/profiles/".$_GET{"profile"};

		if (open($FILE, "+> /tmp/tests_output.8888")) {
			flock($FILE, LOCK_SH);
			my $oldfh = select($FILE);
			$| = 1;
			select($oldfh);
			my $child = fork();
			if (!defined($child)) {
				print $script_begin."setInfo('<font color=\"red\">Cannot create child process!</font>');\n";
				print "ao('Execute output will go here...');".$script_end;
			}
			elsif ($child == 0) {
				unlink("/tmp/tests_status.8888");
				unlink("/tmp/tests_output.8888.old");
				print $FILE "PID: $$";
				my $test_child = fork();
				my $res;
				if (!defined($test_child)) {
					print $script_begin."setInfo('<font color=\"red\">Cannot create test child process!</font>');\n";
					print "ao('Execute output will go here...');".$script_end;
				}
				elsif ($test_child == 0) {
					print $FILE ", CHLD_PID: $$\n";
					close($FILE);
					exec($CONFIG{"TESTS_DIR"}."/lsb-tef.pl -f \"$profile_path\" --results-dir=\"".$CONFIG{"RESULTS_DIR"}."\" -op >> /tmp/tests_output.8888 2>&1");
				}
				else {
					close($FILE);
					waitpid($test_child, 0);
					$res = $?;
				}
				open(FILE_STAT, "> /tmp/tests_status.8888");
				print FILE_STAT $res."\n";
				if ($res == 0) {
					if (open(FILE_HIST, $CONFIG{"RESULTS_DIR"}."/HISTORY")) {
						flock(FILE_HIST, LOCK_SH);
						my $tmp;
						while(<FILE_HIST>) {
							$tmp = $_;
						}
						flock(FILE_HIST, LOCK_UN);
						close(FILE_HIST);
						$tmp =~ s/[\r\n]//g;
						my $path = $CONFIG{"RESULTS_DIR"}."/$tmp";
						if (-d $path) {
							print FILE_STAT "$tmp\n";
						}
						else {
							print STDERR "Wrong directory specified in the HISTORY file!\n";
						}
					}
					else {
						print STDERR "Cannot open HISTORY file!\n";
					}
				}
				close(FILE_STAT);
				exit;
			}
			else {
				select(undef, undef, undef, 0.5);
				seek($FILE, 0, 0);
				my $kill_PID = <$FILE>;
				$kill_PID =~ s/.*CHLD_PID: (\d+).*/$1/g;
				$kill_PID =~ s/[\r\n]//g;

				do_update_output($child, $kill_PID, "/tmp/tests_output.8888", 1, $FILE);
			}
		}
		else {
			print $script_begin."setInfo('<font color=\"red\">Cannot open output file!</font>');\n";
			print "ao('Execute output will go here...');".$script_end;
		}
	}
	else {
		print $script_begin."setInfo('Nothing started');\n";
		print "ao('Execute output will go here...');".$script_end;
	}
}

print $script_begin."execFinished=true;\nrefresh(true);".$script_end;
print "</body>\n</html>\n";
